#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

import re
import sys
import unittest
from math import sqrt


class TransformationMatrix(object):
    """
    Класс подготовки матрицы трансформации из данных, полученных от xrandr и xinput_calibrator
    """

    def __init__(self):
        self.rotate_type = "normal"
        self.rotate_mods = {"normal":   [1.,  0.,  0., 0.,  1.,  0., 0., 0., 1.],
                            "left":     [0.,  -1., 1., 1.,  0.,  0., 0., 0., 1.],
                            "right":    [0.,  1.,  0., -1., 0.,  1., 0., 0., 1.],
                            "inverted": [-1., 0.,  1., 0.,  -1., 1., 0., 0., 1.]}
        self.width_mod = 1.
        self.height_mod = 1.
        self.calibration_mod = [0., 0., 0., 0., 0., 0., 0., 0., 1.]

    def set_rotate_type(self, rotate_type):
        if rotate_type in self.rotate_mods:
            self.rotate_type = rotate_type

    def update_scale_mods(self, out_xrandr):
        # Из вывода xrandr берем размеры скрина, размеры главного монитора и вычисляем модификаторы масштабирования
        screen_w = int(re.search(r".*Screen.*current\s([0-9]*)\sx.*", out_xrandr).group(1))
        screen_h = int(re.search(r".*Screen.*current\s[0-9]*\sx\s([0-9]*).*", out_xrandr).group(1))
        primary_w = int(re.search(r".*primary\s([0-9]*)x.*", out_xrandr).group(1))
        primary_h = int(re.search(r".*primary\s[0-9]*x([0-9]+)+.*", out_xrandr).group(1))
        self.width_mod = float(primary_w) / float(screen_w)
        self.height_mod = float(primary_h) / float(screen_h)

    def update_calibration_mod(self, out_xinput_calibrator):
        # Из вывода xinput_calibrator получаем результаты калибровки и вычисляем модификаторы калибровки
        w_tot = int(re.search(r".*max_x=([0-9]*).*", out_xinput_calibrator).group(1))
        h_tot = int(re.search(r".*max_y=([0-9]*).*", out_xinput_calibrator).group(1))

        match = re.search(r".*Calibration\"\s\"([-0-9 ]*).*", out_xinput_calibrator)
        if match is not None:
            min_x, max_x, min_y, max_y = map(lambda x: int(x), filter(None, match.group(1).split(' ')))
        else:
            min_x = int(re.search(r".*MinX\"\s\"([-0-9]*).*", out_xinput_calibrator).group(1))
            max_x = int(re.search(r".*MaxX\"\s\"([-0-9]*).*", out_xinput_calibrator).group(1))
            min_y = int(re.search(r".*MinY\"\s\"([-0-9]*).*", out_xinput_calibrator).group(1))
            max_y = int(re.search(r".*MaxY\"\s\"([-0-9]*).*", out_xinput_calibrator).group(1))

        w_touch = float(max_x - min_x)
        h_touch = float(max_y - min_y)

        c0 = float(w_tot) / w_touch
        c1 = -1 * float(min_x) / w_touch
        c2 = float(h_tot) / h_touch
        c3 = -1 * float(min_y) / h_touch

        self.calibration_mod[0] = c0
        self.calibration_mod[2] = c1
        self.calibration_mod[4] = c2
        self.calibration_mod[5] = c3

    def get_transformation_matrix(self, default=False):
        calibration_matrix = Matrix(self.calibration_mod.copy())
        rotate_matrix = Matrix(self.rotate_mods[self.rotate_type].copy())

        result = [0, 0, 0, 0, 0, 0, 0, 0, 0]

        if default:
            # Применяем модификаторы поворота
            result = self.rotate_mods[self.rotate_type].copy()
        else:
            # Применяем модификаторы поворота и калибровки
            for row in range(0, rotate_matrix.size):
                for column in range(0, rotate_matrix.size):
                    index = row * rotate_matrix.size + column
                    result[index] = result[index] + multiple(rotate_matrix.row(row), calibration_matrix.column(column))

        # Применяем модификаторы масштабирования
        for i in range(3):
            result[i] *= self.width_mod
            result[i+3] *= self.height_mod

        # Округление
        for i in range(len(result)):
            result[i] = round(result[i], 3)

        return result


class TestTfMatrix(unittest.TestCase):

    def test_update_scale_mods(self):
        screen_w = "3840"
        screen_h = "1080"
        primary_w = "1280"
        primary_h = "1024"
        mod_w = float(primary_w) / float(screen_w)
        mod_h = float(primary_h) / float(screen_h)
        out_xrandr = "Screen 0: minimum 320 x 200, current " + screen_w + " x " + screen_h + ", maximum 8192 x 8192\n" \
                     "VGA-1 connected 1920x1080+0+0 (normal left inverted right x axis y axis) 477mm x 268mm\n" \
                     "   1920x1080     60.00*+\n" \
                     "   1600x1200     60.00\n" \
                     "   1680x1050     59.95\n" \
                     "   1280x1024     75.02    60.02\n" \
                     "HDMI-1 connected primary " + primary_w + "x" + primary_h + "+1920+0 " \
                     "(normal left inverted right x axis y axis) 521mm x 293mm\n" \
                     "   1920x1080     60.00*+  50.00    59.94\n" \
                     "   1680x1050     59.88\n" \
                     "   1600x900      60.00\n" \
                     "   1280x1024     75.02    60.02"

        tr_matrix = TransformationMatrix()
        tr_matrix.update_scale_mods(out_xrandr)
        print(tr_matrix.width_mod)
        print(tr_matrix.height_mod)
        self.assertEqual(mod_w, tr_matrix.width_mod)
        self.assertEqual(mod_h, tr_matrix.height_mod)

    def test_update_calibration_mod(self):
        out_xinput_calibrator = 'Calibrating standard Xorg driver "eGalax Inc. USB TouchController"' \
                                '	current calibration values: min_x=0, max_x=65535 and min_y=0, max_y=65535' \
                                '	If these values are estimated wrong, either supply it manually with the --precalib ' \
                                'option, or run the "get_precalib.sh" script to automatically get it (through HAL).' \
                                '	--> Making the calibration permanent <--' \
                                'copy the snippet below into "/etc/X11/xorg.conf.d/99-calibration.conf" ' \
                                '(/usr/share/X11/xorg.conf.d/ in some distro"s)' \
                                'Section "InputClass"' \
                                '	Identifier	"calibration"' \
                                '	MatchProduct	"eGalax Inc. USB TouchController"' \
                                '	Option	"MinX"	"-171"' \
                                '	Option	"MaxX"	"66047"' \
                                '	Option	"MinY"	"-728"' \
                                '	Option	"MaxY"	"63351"' \
                                '	Option	"SwapXY"	"0" # unless it was already set to 1' \
                                '	Option	"InvertX"	"0"  # unless it was already set' \
                                '	Option	"InvertY"	"0"  # unless it was already set' \
                                'EndSection'
        value1 = 65535. / (66047. - (-171.))
        value2 = (-1 * (-171.) / (66047. - (-171.)))
        value3 = 65535. / (63351. - (-728.))
        value4 = (-1 * (-728.) / (63351. - (-728.)))

        expected_mod = [value1, 0., value2, 0., value3, value4, 0., 0., 1.]

        tr_matrix = TransformationMatrix()
        tr_matrix.update_calibration_mod(out_xinput_calibrator)

        self.assertEqual(expected_mod, tr_matrix.calibration_mod)

    def test_get_transformation_matrix_default(self):
        tr_matrix = TransformationMatrix()

        for rotate_type in tr_matrix.rotate_mods:
            with self.subTest(i=rotate_type):
                tr_matrix.set_rotate_type(rotate_type)
                expected_matrix = tr_matrix.rotate_mods[rotate_type].copy()

                result = tr_matrix.get_transformation_matrix(True)

                self.assertEqual(expected_matrix, result)


class Matrix(object):
    """
    Класс для получения строк и столбцов квадратной матрицы
    """

    def __init__(self, array):
        self.array = array
        self.size = int(sqrt(len(array)))

    def row(self, index):
        return self.array[(index * self.size):(index * self.size + self.size)]

    def column(self, index):
        result = []
        for i in range(0, int(len(self.array) / self.size)):
            result.append(self.array[index + i * self.size])

        return result


def multiple(array1, array2):
    result = 0.0
    for i in range(0, len(array1)):
        result += array1[i] * array2[i]
    return result


def list_to_str(value):
    result = []
    for i in range(0, len(value)):
        result.append(str(value[i]))
    return " ".join(result)


if __name__ == "__main__":
    """
    Утилита для получения матрицы трансформации при калибровке
    """
    # Вызов без аргументов запускает тесты
    if len(sys.argv) == 1:
        unittest.main()
        exit(0)

    transformation_matrix = TransformationMatrix()
    # Устанавливаем поворот
    transformation_matrix.set_rotate_type(sys.argv[1])
    # Обрабатываем вывод xrandr
    transformation_matrix.update_scale_mods(sys.argv[2])

    # Получили только поворот и масштаб
    if len(sys.argv) == 3:
        print(list_to_str(transformation_matrix.get_transformation_matrix(default=True)))
        exit(0)

    # Обрабатываем вывод xinput_calibrator
    transformation_matrix.update_calibration_mod(sys.argv[3])

    print(list_to_str(transformation_matrix.get_transformation_matrix()))
